// Analyse des FS20 S3 Sender Impulsstromes Richtung HF-Teil  - Code-Sniffer mit Arduino
// 
// Ein Arduino Nano liest die Impulsfolge des Digitalteils eines FS20 S4M ein, der auf die Schaltsteckdose 
//   des Garagentorantriebs eingelernt ist.
// Dieser impulsstrom wird anschliessend im AtTiny85 des FS20S3-Ersatz fest einprogrammiert.

#include <avr/interrupt.h>

// defines fuer Arduino Nano
#define FSin 2							// hier kommt das serielle Signal vom Digitalteil an

// globale Variablen
volatile unsigned long micro = 0;
volatile unsigned long FS20Micros = 0;	//Time to check DCC Data "1" or "0"
volatile unsigned long TelegrammTimer = 0;

volatile byte data[8];					//eingelesene Bit-Werte, 8 Byte == 64 bit
byte datalength = 0;					//Position in data wo gerade der nchste Wert hinkommt
byte countbit = 0;
byte countbyte = 0;
volatile bool BitSet = false;
volatile bool TelegrammFertig = false;

					  
// Interrruptroutine fuer Impulse
//--------------------------------------------------------------------------------
void FS20data() 
{
	micro = micros();							// aktuelle Zeit merken

	//digitalWrite(13, !digitalRead(13));				// debug
	digitalWrite(12,1);							// debug

	if(micro > TelegrammTimer + 5000)			// das letzte Bit ist laenger als 5 ms her -> neues Telegramm
	{											// also setze alle relevanten Zaehler auf Startbedingung
		datalength = 0;							// Position im double an der die Bitwerte gespeichert werden.
		//BitVal = 0xFF;
		TelegrammFertig = false;	

		//digitalWrite(12,0);							// debug
	}

	if (digitalRead(FSin) == HIGH)				// Signal noch HIGH?
	{											// ja, also steigende Flanke
		FS20Micros = micro;						// Zeit merken
	}
	else 
	{											// nein, also fallende Flanke -> auswerten
		countbit = datalength % 8;				// countbit laeuft von 0..7
		countbyte = datalength / 8;				// countbyte laeuft von 0..7

		if (micro <= (FS20Micros + 500)) {      // 0-Bit gelesen; kuerzer als 500us
			//digitalWrite(12,0);						// debug
			bitWrite(data[countbyte], 7 - countbit, 0);	//Speichert das gelesene Bit (startet beim least significant == rightmost bit)
		}
		else 
		{										// 1-Bit gelesen; laenger als 500us
			//digitalWrite(12,1);						// debug
			bitWrite(data[countbyte], 7 - countbit, 1);	//Speichert das gelesene Bit (startet beim least significant == rightmost bit)
		}

		datalength += 1;						// Abzhlen der Bits, also max 64, Telegramm hat 59 Bit
		if (datalength > 59)
		{
			TelegrammFertig = true;
			digitalWrite(12,0);							// debug
		}

		TelegrammTimer = micro;					// es war ein Bit da, also Telegramm unterwegs
	}
}
//--------------------------------------------------------------------------------

void setup()
{
	// Serielle SS initialisieren
	Serial.begin(115200);

	// Datenrichtung setzen
	pinMode(FSin, INPUT_PULLUP);
	pinMode(13, OUTPUT);
	pinMode(12, OUTPUT);

	attachInterrupt(digitalPinToInterrupt(FSin), FS20data, CHANGE);				//ISR fr den Dateneingang

	FS20Micros = micros();
}

void loop()
{
	// empfangenes FS20 Telegramm decodieren
	FS20updatedata();
}

//--------------------------------------------------------------------------------
//need to update inside each loop
void FS20updatedata() {
	
	//digitalWrite(13, !digitalRead(13));		// debug
	{
		if (TelegrammFertig == true)			// wurde ein Bit erkannt?
		{
			cli();								// IRQs sperren, sonst hauen die Wiederholungen in die serielle Ausgabe
			if (datalength >= 59) 				// End-Bit gelesen.
			{
				for (char i = 0; i < 59; i++)
				{
					Serial.print(i / 10);
				} 
				Serial.println();
				for (char i = 0; i < 59; i++)
				{
					Serial.print(i% 10); 
				}
				Serial.println();
				Serial.println();
				for (char i = 0; i < 60; i++)
				{
					countbit = i % 8;			// countbit laeuft von 0..7
					countbyte = i / 8;			// countbyte laeuft von 0..7

					Serial.print((bitRead(data[countbyte], 7 - countbit)) ? "1" : "0");
				}
				Serial.println();
			}
			sei();								// IRQs wieder freigeben
			TelegrammFertig = false;
		}
	}
}


// bisher nicht benoetigte Routinen
// ---------------------------------------------------------------------------------------------------------------------
void FS20auswertung()
{
	// FS20 Telegramm allgemein
	// Synchronisation
	// Jede bertragung zu Empfngern wird mit dieser Synchronisationssequenz eingeleitet:
	// 0000000000001 (12 x 0, 1 x 1)
	// Paritt: even / Gerade
	// EOT : 0
	// Damit das letzte Datenpaket richtig ausgewertet werden kann, muss ein zustzliches Daten-Bit bertragen werden.
	// Quersumme: 8bit-Summe aus 0x06, Hauscode-High-Byte HC1, Hauscode-Low-Byte HC2, Adresse und Befehl
	//
	// Komplette bertragung - Erweiterungsbit=0: (HC1/2 -Hauscode)
	// Synchr	HC1		Parity	HC2		Parity	Adresse	Parity	Befehl	Parity	Quersumme	Parity	EOT
	// 13 bit	8 bit	1 bit	8 bit	1 bit	8 bit	1 bit	8 bit	1 bit	8 bit		1 bit	1 bit	gesamt: 59 bit
	//
	// Komplette bertragung - Erweiterungsbit=1: (HC1/2 -Hauscode)  -- nicht verwendet bei FS20 S3 --
	// Synchr	HC1		Parity	HC2		Parity	Adresse	Parity	Befehl	Parity	Erweiterung	Parity	Quersumme	Parity	EOT 
	// 13 bit	8 bit	1 bit	8 bit	1 bit	8 bit	1 bit	8 bit	1 bit	8 bit		1 bit	8 bit		1 bit	1 bit	gesamt: 68 bit




}

// Returns 1 (ODD) or 0 (EVEN) parity
int parity(unsigned char x)
{
	x = x ^ x >> 4;
	x = x ^ x >> 2;
	x = x ^ x >> 1;
	return x & 1;
}
// ---------------------------------------------------------------------------------------------------------------------
